"""
This package adds distortion to matrices.
"""
import numpy as np


def add_gaussian_noise(mat, mean, std):
    '''
    Adds gaussian noise to the matrix

    Args:
        mat: 2d array, shape (n,n)
        mean: gaussian mean
        std: gaussian std
    Returns:
        Processed matrix
    '''
    noise_filter = std * np.random.randn(mat.shape[0], mat.shape[1]) + mean
    return mat + noise_filter


def shift(mat, n_shift):
    '''
    Shifts the matrix by rolling it along the diagonal

    Args:
        mat: 2d array, shape (n,n)
        n_shift: number to roll
    Returns:
        Shifted matrix
    '''
    if mat.shape[0] != mat.shape[1]:
        raise ValueError("mat needs to be square matrix.")
    shifted_mat = np.roll(mat, n_shift, axis=0)
    shifted_mat = np.roll(shifted_mat, n_shift, axis=1)
    return shifted_mat



"""
This package contains the algorithm in 'A feasible method for optimizationwith orthogonality constraints' by Zaiwen Wen and Wotao Yin.

We have reinterpreted the algorithm from Matlab by Wen & Yin to Python.
"""

import numpy as np


def opt_mani_mulit_ball_gbb(x,
                            fun,
                            args,
                            xtol=1e-6,
                            ftol=1e-12,
                            gtol=1e-6,
                            rho=1e-4,
                            eta=0.1,
                            gamma=0.85,
                            tau=1e-3,
                            nt=5,
                            mxitr=1000,
                            record=0):
    """
    Line search algorithm for optimization on manifold
    Reinterpreted directly from Zaiwen Wen and Wotao Yin's
    Matlab implementation of their paper on
    'A feasible method for optimizationwith orthogonality constraints'

    Args:
        x: Numpy array where each column lies on the unit
            sphere ||x_i||_2 = 1
        fun: Function that returns the objective function value
            and its gradient.
            Params: [x, args]
            Returns: [f, g]
        args: args to be used in fun
        kwargs: Options
            record = 0, no print out
            mxitr       max number of iterations
            xtol        stop control for ||X_k - X_{k-1}||
            gtol        stop control for the projected gradient
            ftol        stop control for abs(F_k - F_{k-1})/(1+|F_{k-1}|) usually, max{xtol, gtol} > ftol
    Returns:
        (x, g, out)

        x: solution

        g: gradient of x

        Out: output information

    """
    out = {}

    crit = np.ones((mxitr, 3))

    # Normalize x
    (n, p) = x.shape
    nrmx = np.sum(x * x, axis=0)
    if np.linalg.norm(nrmx) > 1e-8:
        x = np.divide(x, np.sqrt(nrmx))

    # Initial function value and gradient
    f, g = fun(x, args)
    xtg = np.sum(x * g, axis=0)
    gg = np.sum(g * g, axis=0)
    xx = np.sum(x * x, axis=0)
    xxgg = xx * gg
    dtX = x * xtg - g
    nrmG = np.linalg.norm(dtX, 'fro')

    Q = 1
    Cval = f
    tau_orig = tau
    tau = tau_orig
    if (record >= 1):
        print('----------- Gradient Method with Line search -----------')
        print(
            '{:4} \t {:10} \t {:10} \t  {:10} \t {:5} \t {:9} \t {:7}'.format(
                'Iter', 'tau', 'f(X)', 'nrmG', 'Exit', 'funcCount', 'ls-Iter'))
        print(
            '{:4} \t {:10} \t {:10} \t  {:10} \t {:5} \t {:9} \t {:7}'.format(
                0, 0, f, 0, 0, 0))
    if record == 10:
        out['fvec'] = f

    # Main Iteration
    for itr in range(mxitr):
        xp = x
        fp = f
        gp = g
        dtXP = dtX

        nls = 1
        deriv = rho * nrmG**2
        while True:
            tau2 = tau / 2
            beta = 1 + tau2**2 * (-xtg**2 + xxgg)
            a1 = ((1 + tau2 * xtg)**2 - tau2**2 * xxgg) / beta
            a2 = -tau * xx / beta
            x = xp * a1 + gp * a2

            f, g = fun(x, args)
            out['nfe'] = out['nfe'] + 1 if 'nfe' in out else 1

            if f <= Cval - tau * deriv or nls >= 5:
                break
            tau = eta * tau
            nls = nls + 1

        if record == 10:
            out['fvec'] = [out['fvec'], f] if 'fvec' in out else [f]

        # Recalculate
        xtg = np.sum(x * g, axis=0)
        gg = np.sum(g * g, axis=0)
        xx = np.sum(x * x, axis=0)
        xxgg = xx * gg
        dtX = x * xtg - g
        nrmG = np.linalg.norm(dtX, 'fro')
        s = x - xp
        XDiff = np.linalg.norm(s, 'fro') / np.sqrt(n)
        FDiff = abs(fp - f) / (abs(fp) + 1)

        if record >= 1:
            print(('{:4d} \t {:3.2e} \t {:7.6e} \t {:3.2e} \t {:3.2e}'
                   '\t {:3.2e} \t {:2d}\n').format(itr, tau, f, nrmG, XDiff,
                                                   FDiff, nls))

        crit[itr, :] = [nrmG, XDiff, FDiff]
        mcrit = np.mean(crit[itr - min(nt, itr):itr + 1, :], axis=0)

        if ((XDiff < xtol and FDiff < ftol) or nrmG < gtol
                or np.all(mcrit[1:] < 10 * np.array([xtol, ftol]))):
            out['msg'] = 'converge'
            break

        y = dtX - dtXP
        sy = np.sum(s * y)
        sy = abs(sy)
        tau = tau_orig
        if sy > 0:
            if np.mod(itr, 2) == 0:
                tau = np.sum(s * s) / sy
            else:
                tau = sy / np.sum(y * y)

            # Safeguarding on tau
            tau = max(min(tau, 1e20), 1e-20)

        Qp = Q
        Q = gamma * Qp + 1
        Cval = (gamma * Qp * Cval + f) / Q

    if itr >= mxitr:
        out['msg'] = 'exceed max iteration'

    out['feasi'] = np.linalg.norm(np.sum(x * x, axis=0) - 1)
    if out['feasi'] > 1e-14:
        nrmx = np.sum(x * x, axis=0)
        x = np.divide(x, np.sqrt(nrmx))
        f, g = fun(x, args)
        out['nfe'] = out['nfe'] + 1
        out['feasi'] = np.linalg.norm(np.sum(x * x, axis=0) - 1)

    out['nrmG'] = nrmG
    out['fval'] = f
    out['itr'] = itr
    return x, g, out


def maxcut_quad(V, C):
    """
    Maxcut function to compute objective function value and gradient

    maxcut SDP:
    X is n by n matrix
    max Tr(C*X), s.t., X_ii = 1, X psd

    Args:
        V: ndarray, Low rank model X = V' * V;
        C: ndarray, Objective matrix to compute maxcut

    Returns:
        (f, g)

        f: float, objective funciton value

        g: ndarray, gradient
    """
    # Only taking first arg
    g = 2 * np.matmul(V, C)
    f = np.sum(g * V) / 2
    return f, g


def cutnorm_quad(V, C):
    """
    Cutnorm function to compute objective function value and gradient

    Args:
        V: ndarray, Low rank model X = V' * V;
        C: ndarray, Objective matrix to compute maxcut

    Returns:
        (f, g)

        f: float, objective funciton value

        g: ndarray, gradient
    """
    n = len(C)
    Vs = V[:, n:]
    Us = V[:, :n]

    g = 2 * np.c_[np.matmul(Vs, C.T), np.matmul(Us, C)]
    f = (np.sum(g[:, :n] * Us) + np.sum(g[:, n:] * Vs)) / 2
    return f, g



"""
This package generates stochastic matrices.

Currently supported models are general Stochastic Block Models, Erdos Renyi, and Autoregressive Models.
"""
import numpy as np


def sbm(community_sizes, prob_mat, symmetric=True):
    '''
    Generates a stochastic block matrix

    Community_sizes indicate the size of each community and the
    probability matrix indicate the probability that a 1 will be
    generated for each element within the community.

    Args:
        community_sizes: 1d array, shape (n) sizes of community
        prob_mat: 2d array, shape (n,n) probability of edges for each community
        symmetric: boolean, if true, the function will output a symmetric matrix
    Returns:
        stochastic block matrix, 2d array, shape depending on community sizes
    '''
    community_sizes = np.array(community_sizes)
    prob_mat = np.array(prob_mat)
    if prob_mat.shape[0] != prob_mat.shape[1]:
        raise ValueError("prob_mat needs to be square matrix.")
    if not np.all(community_sizes > 0):
        raise ValueError("Each community size in community_sizes "
                         "needs to be greater than 1.")
    if len(community_sizes) != len(prob_mat):
        raise ValueError("community_sizes needs to be of size n if "
                         "prob_mat is nxn")
    if not (np.all(prob_mat >= 0) and np.all(prob_mat <= 1)):
        raise ValueError("Needs to be a valid probability matrix.")

    n = np.sum(community_sizes)
    sbm = np.zeros((n, n))
    for i, size_i in enumerate(community_sizes):
        for j, size_j in enumerate(community_sizes):
            prev_sum_i = np.sum(community_sizes[:i])
            prev_sum_j = np.sum(community_sizes[:j])
            prob = prob_mat[i][j]
            sample = np.random.choice(
                2, size=size_i * size_j, p=[1 - prob, prob]).reshape((size_i,
                                                                      size_j))
            sbm[prev_sum_i:prev_sum_i + size_i, prev_sum_j:
                prev_sum_j + size_j] = sample

    # Discard lower triangular and reflect upper triangular
    if symmetric:
        sbm = make_symmetric_triu(sbm)
    return sbm


def sbm_prob(community_sizes, prob_mat):
    '''
    Generates a matrix indicating the underlying probability
    that gives rise to a stochastic block matrix

    Args:
        community_sizes: 1d array, shape (n) sizes of community
        prob_mat: 2d array, shape (n,n) probability of edges for each community
    Returns:
        probabilities of a stochastic block matrix,
        2d array, shape depending on community sizes
    '''
    community_sizes = np.array(community_sizes)
    prob_mat = np.array(prob_mat)
    if prob_mat.shape[0] != prob_mat.shape[1]:
        raise ValueError("prob_mat needs to be square matrix.")
    if not np.all(community_sizes > 0):
        raise ValueError("Each community size in community_sizes "
                         "needs to be greater than 1.")
    if len(community_sizes) != len(prob_mat):
        raise ValueError("community_sizes needs to be of size n if "
                         "prob_mat is nxn")
    if not (np.all(prob_mat >= 0) and np.all(prob_mat <= 1)):
        raise ValueError("Needs to be a valid probability matrix.")

    n = np.sum(community_sizes)
    sbm = np.zeros((n, n))
    for i, size_i in enumerate(community_sizes):
        for j, size_j in enumerate(community_sizes):
            prev_sum_i = np.sum(community_sizes[:i])
            prev_sum_j = np.sum(community_sizes[:j])
            prob = prob_mat[i][j]
            prob_block = prob * np.ones((size_i, size_j))
            sbm[prev_sum_i:prev_sum_i + size_i, prev_sum_j:
                prev_sum_j + size_j] = prob_block

    return sbm


def sbm_autoregressive(community_sizes, prob_list, symmetric=True):
    '''
    Generates an autoregressive SBM

    An autoregressive SBM has edge probability according to the prob_list
    on the diagonal but (prob_list[i] * prob_list[j])**(abs(i - j))
    for the off-diagonal blocks entries.

    This idea is similar to the autoregressive models

    Args:
        community_sizes: 1d array, shape (n) sizes of community
        prob_list: 1d array, shape (n), where n is the number of diagonal blocks
        symmetric: boolean, if true, the function will output a symmetric matrix
    Returns:
        An autoregressive SBM, 2d array, shape depending on community sizes
    '''
    prob_matrix = _sbm_autoregressive_gen_prob_matrix(prob_list)
    return sbm(community_sizes, prob_matrix, symmetric)


def sbm_autoregressive_prob(community_sizes, prob_list):
    '''
    Generates the underlying probability matrix thatgives
    rise to the autoregressive SBM

    Args:
        community_sizes: 1d array, shape (n) sizes of community
        prob_list: 1d array, shape (n), where n is the number of diagonal blocks
    Returns:
        A probability matrix for an autoregressive SBM,
        2d array, shape depending on community sizes
    '''
    prob_matrix = _sbm_autoregressive_gen_prob_matrix(prob_list)
    return sbm_prob(community_sizes, prob_matrix)


def _sbm_autoregressive_gen_prob_matrix(prob_list):
    '''
    Generates the probability block matrix for an autoregressive SBM

    Each element of the probaiblity block matrix represents the
    probability of an edge within the block. This is of shape (n,n)

    This is different than sbm_autoregressive_prob in that this is
    shape (n,n) whereas the shape of sbm_autoregressive_prob depends
    on the community sizes.

    Args:
        prob_list: 1d array, shape (n), where n is the number of diagonal blocks
    Returns:
        A probability matrix for an autoregressive SBM,
        2d array, shape (n,n)
    '''
    n_probs = len(prob_list)
    prob_matrix = np.zeros((n_probs, n_probs))
    for i in range(n_probs):
        for j in range(n_probs):
            if i == j:
                prob_matrix[i, i] = prob_list[i]
            else:
                prob_matrix[i, j] = (prob_list[i] * prob_list[j])**(abs(i - j))
    return prob_matrix


def make_symmetric_triu(mat):
    '''
    Makes the matrix symmetric upper triangular

    Args:
        mat: 2d array, shape (n,n)
    Returns:
        upper triangular symmetric matrix of the input
        2d array, shape (n,n)
    '''
    mat = np.triu(mat)
    mat = np.maximum(mat, mat.T)
    return mat


def erdos_renyi(n, p, symmetric=True):
    '''
    Generates Erdos Renyi random graph size n with
    probability p

    Args:
        n: int, size of the output matrix
        p: float, edge probability
        symmetric: boolean, if true, the function will output a symmetric matrix
    Returns:
        Erdos Renyi random graph matrix
        2d array, shape (n,n)
    '''
    return sbm([n], [[p]], symmetric)



"""
This package computes the lp distances given two matrices.

Matrices of different dimension and weights will be rescaled and reweighted using the same method in Cutnorm package.
"""
import numpy as np
from ..compute import _compute_C_weighted, _compute_C_eqdim_unweighted, _compute_C_uneqdim_unweighted


def compute_lp_distance(p, A, B, w1=None, w2=None):
    # Input checking
    if type(A) is not np.ndarray:
        A = np.array(A)
    if type(B) is not np.ndarray:
        B = np.array(B)
    if A.ndim != 2 or B.ndim != 2:
        raise ValueError("A and B must be 2D matrices")
    n, n2 = np.shape(A)
    m, m2 = np.shape(B)
    if n != n2 or m != m2:
        raise ValueError("A and B must be square matrices")
    if (w1 is None and w2 is not None) or (w1 is not None and w2 is None):
        raise ValueError("Weight vectors required for both matrices")
    if (w1 is not None and w2 is not None and (n != len(w1) or m != len(w2))):
        raise ValueError("Weight vectors need to have the same lenght "
                         "as the first dimension of the corresponding "
                         "matrices")

    if w1 is not None:
        w, C = _compute_C_weighted(A, B, w1, w2)
    else:
        if n == m:
            w, C = _compute_C_eqdim_unweighted(A, B)
        else:
            w, C = _compute_C_uneqdim_unweighted(A, B)

    if p == np.inf:
        return np.linalg.norm(C.flatten(), ord=p) * (n**2)
    return np.linalg.norm(C.flatten(), ord=p) * (n**2) / (n**(2 / p))



"""
This package computes the Cutnorm.
"""
import time
import math
import numpy as np
from .OptManiMulitBallGBB import opt_mani_mulit_ball_gbb, cutnorm_quad


def compute_cutnorm(A,
                    B,
                    w1=None,
                    w2=None,
                    max_round_iter=100,
                    logn_lowrank=False,
                    extra_info=False):
    """
    Computes the cutnorm of the differences between the two matrices

    Args:
        A: ndarray, (n, n) matrix
        B: ndarray, (m, m) matrix
        w1: ndarray, (n, 1) array of weights for A
        w2: ndarray, (m, 1) array of weights for B
        max_round_iter: int, number of iterations for gaussian rounding
        logn_lowrank: boolean to toggle log2(n) low rank approximation
        extra_info: boolean, generate extra computational information
    Returns:
        (cutnorm_round, cutnorm_sdp, info)

        cutnorm_round: objective function value from gaussian rounding

        cutnorm_sdp: objective function value from sdp solution

        info: dictionary containing computational information
            Computational information from OptManiMulitBallGBB:
                sdp_augm_n: dimension of augmented matrix
                sdp_relax_rank_p: rank
                sdp_tsolve: computation time
                sdp_itr, sdp_nfe, sdp_feasi, sdp_fval, sdp_g, sdp_nrmG: information from OptManiMulitBallGBB
            Computational information from gaussian rounding:
                round_tsolve: computation time for rounding
                round_approx_list: list of rounded objf values
                round_uis_list: list of uis
                round_vjs_list: list of vjs
                round_uis_opt: optimum uis
                round_vjs_opt: optimum vjs

            Computational information from processing the difference:
                weight_of_C: weight vector of C, the difference matrix

            Cutnorm information:
               cutnorm_sets (S,T): vectors of cutnorm
    Raises:
        ValueError: if A and B are of wrong dimension, or if weight vectors
            does not match the corresponding A and B matrices
    """
    # Input checking
    if not isinstance(A, np.ndarray):
        A = np.array(A)
    if not isinstance(B, np.ndarray):
        B = np.array(B)
    if A.ndim != 2 or B.ndim != 2:
        raise ValueError("A and B must be 2D matrices")
    n, n2 = np.shape(A)
    m, m2 = np.shape(B)
    if n != n2 or m != m2:
        raise ValueError("A and B must be square matrices")
    if (w1 is None and w2 is not None) or (w1 is not None and w2 is None):
        raise ValueError("Weight vectors required for both matrices")
    if (w1 is not None and w2 is not None and (n != len(w1) or m != len(w2))):
        raise ValueError("Weight vectors need to have the same lenght "
                         "as the first dimension of the corresponding "
                         "matrices")

    if w1 is not None:
        w, C = _compute_C_weighted(A, B, w1, w2)
    else:
        if n == m:
            w, C = _compute_C_eqdim_unweighted(A, B)
        else:
            w, C = _compute_C_uneqdim_unweighted(A, B)

    cutnorm_round, cutnorm_sdp, info = _compute_cutnorm(
        C, max_round_iter, logn_lowrank, extra_info)

    # Add weight vector into extra_info info
    info['weight_of_C'] = w

    return cutnorm_round, cutnorm_sdp, info


def gaussian_round(U,
                   V,
                   C,
                   max_round_iter,
                   logn_lowrank=False,
                   extra_info=False):
    '''
    Gaussian Rounding for Cutnorm

    The algorithm picks a random standard multivariate gaussian vector
    w in R^p and computes the rounded solution based on sgn(w \\dot ui).

    Adopted from David Koslicki's cutnorm rounding code
    https://github.com/dkoslicki/CutNorm
    and Peter Diao's modifications

    Args:
        U: ndarray, (p, n) shaped matrices of relaxed solutions
        V: ndarray, (p, n) shaped matrices of relaxed solutions
        C: ndarray, original (n, n) shaped matrix to compute cutnorm
        max_round_iter: maximum number of rounding operations
        logn_lowrank: boolean to toggle log2(n) low rank approximation
        extra_info: boolean, generate extra computational information
    Returns:
        (approx_opt, uis_opt, vjs_opt, round_info)

        approx_opt: approximated objective function value

        uis_opt: rounded u vector

        vis_opt: rounded v vector

        round_info: information for rounding operation
    '''
    (p, n) = U.shape
    approx_opt = 0
    uis_opt = np.zeros(n)
    vjs_opt = np.zeros(n)
    G = np.random.randn(max_round_iter, p)

    # Computational information
    round_info = {}
    if extra_info:
        approx_list = np.zeros(max_round_iter)
        uis_list = np.zeros((max_round_iter, n))
        vjs_list = np.zeros((max_round_iter, n))

    # Decomposition for low rank
    if logn_lowrank:
        C_U, C_s, C_V = np.linalg.svd(C)
        low_rank = int(np.log2(n))

    for i in range(max_round_iter):
        uis = np.sign(np.matmul(G[i], U))
        vjs = np.sign(np.matmul(G[i], V))

        # Rounding
        if logn_lowrank:
            C_U_low_filtered_sum = np.sum(
                C_U[:, :low_rank] * uis[:, np.newaxis], axis=0)
            C_V_low_filtered_sum = np.sum(
                np.dot(np.diag(C_s[:low_rank]), C_V[:low_rank] * vjs), axis=1)
            approx = np.abs(
                np.sum(C_U_low_filtered_sum * C_V_low_filtered_sum))
        else:
            approx = np.abs(np.sum(C * np.outer(uis, vjs)))

        if approx > approx_opt:
            approx_opt = approx
            uis_opt = uis
            vjs_opt = vjs

        if extra_info:
            approx_list[i] = approx / 4.
            uis_list[i] = uis
            vjs_list[i] = vjs

    # Cutnorm is 1/4 of infinity norm
    approx_opt = approx_opt / 4.

    if extra_info:
        round_info = {
            "round_approx_list": approx_list,
            "round_uis_list": uis_list,
            "round_vjs_list": vjs_list,
            "round_uis_opt": uis_opt,
            "round_vjs_opt": vjs_opt
        }

    return approx_opt, uis_opt, vjs_opt, round_info


def cutnorm_sets(uis, vjs):
    """
    Generates the cutnorm sets from the rounded SDP solutions

    Args:
        uis: ndarray, (n+1, ) shaped array of rounded +- 1 solution
        vis: ndarray, (n+1, ) shaped array of rounded +- 1 solution
    Returns:
        (S, T) Reconstructed S and T sets that are {1, 0}^n

        S: Cutnorm set axis = 0

        T: Cutnorm set axis = 1
    """
    S = -1 * uis[-1] * uis[:-1]
    T = -1 * vjs[-1] * vjs[:-1]

    S = (S + 1) / 2
    T = (T + 1) / 2
    return S, T


def _compute_cutnorm(C, max_round_iter, logn_lowrank=False, extra_info=False):
    """
    Computes the cutnorm of square matrix C

    Args:
        C: ndarray, (n, n) matrix
        max_round_iter: int, maximum rounding iterations
        logn_lowrank: boolean to toggle log2(n) low rank approximation
        extra_info: boolean, extra computational information generation
    Returns:
        (cutnorm_round, cutnorm_sdp, info)

        cutnorm_round: objective function value from gaussian rounding

        cutnorm_sdp: objective function value from sdp solution

        info: dictionary containing computational information
            Computational information from OptManiMulitBallGBB:
                sdp_augm_n: dimension of augmented matrix
                sdp_relax_rank_p: rank
                sdp_tsolve: computation time
                sdp_itr, sdp_nfe, sdp_feasi, sdp_fval, sdp_g, sdp_nrmG: information from OptManiMulitBallGBB
            Computational information from gaussian rounding:
               round_tsolve: computation time for rounding
               round_approx_list: list of rounded objf values
               round_uis_list: list of uis
               round_vjs_list: list of vjs
               round_uis_opt: optimum uis
               round_vjs_opt: optimum vjs

            Cutnorm information:
               cutnorm_sets (S,T): vectors of cutnorm
    """
    n1 = len(C)
    C_col_sum = np.sum(C, axis=0)
    C_row_sum = np.sum(C, axis=1)
    C_tot = np.sum(C_col_sum)
    # Transformation to preserve cutnorm and
    # enforces infinity one norm = 4*cutnorm
    C = np.c_[C, -1.0 * C_row_sum]
    C = np.r_[C, [np.concatenate((-1.0 * C_col_sum, [C_tot]))]]

    # Modify rank estimation for SDP relaxation
    p = int(max(min(round(np.sqrt(2 * n1) / 2), 100), 1))

    # Dim for augmented matrix for SDP
    n2 = 2 * n1 + 2

    # Initial point normalized
    x0 = np.random.randn(p, n2)
    nrmx0 = np.sum(x0 * x0, axis=0)
    x0 = np.divide(x0, np.sqrt(nrmx0))

    tic_sdp = time.time()
    x, g, out = opt_mani_mulit_ball_gbb(
        x0,
        cutnorm_quad,
        C,
        record=0,
        mxitr=600,
        gtol=1e-8,
        xtol=1e-8,
        ftol=1e-10,
        tau=1e-3)
    toc_sdp = time.time()
    tsolve_sdp = toc_sdp - tic_sdp

    # SDP upper bound approximation
    U = x[:, :n2 // 2]
    V = x[:, n2 // 2:]
    cutnorm_sdp = np.abs(np.sum(C * np.matmul(U.T, V))) / 4.0

    # Gaussian Rounding
    tic_round = time.time()
    (cutnorm_round, uis, vjs, round_info) = gaussian_round(
        U, V, C, max_round_iter, logn_lowrank, extra_info)
    toc_round = time.time()
    tsolve_round = toc_round - tic_round

    # Generate cutnorm sets
    (S, T) = cutnorm_sets(uis, vjs)

    info = {
        "cutnorm_sets": (S, T),
        "sdp_tsolve": tsolve_sdp,
        "round_tsolve": tsolve_round
    }
    if extra_info:
        info.update({
            "sdp_augm_n": n2,
            "sdp_relax_rank_p": p,
            "sdp_itr": out['itr'],
            "sdp_nfe": out['nfe'],
            "sdp_feasi": out['feasi'],
            "sdp_fval": out['fval'],
            "sdp_g": g,
            "sdp_nrmG": out['nrmG']
        })
        # Join rounding info
        info.update(round_info)

    return cutnorm_round, cutnorm_sdp, info


def _compute_C_weighted(A, B, w1, w2):
    """
    Generates the difference matrix of the two weighted matricies

    Args:
        A: ndarray, (n, n) matrix
        B: ndarray, (m, m) matrix
        w1: ndarray, (n, 1) array of weights for A
        w2: ndarray, (m, 1) array of weights for B
    Returns:
        C: ndarray, the difference matrix
    """
    v1 = np.hstack((0, np.cumsum(w1)[:-1], 1.))
    v2 = np.hstack((0, np.cumsum(w2)[:-1], 1.))
    v = np.unique(np.hstack((v1, v2)))
    w = np.diff(v)
    n1 = len(w)

    a = np.zeros(n1, dtype=np.int32)
    b = np.zeros(n1, dtype=np.int32)
    for i in range(n1 - 1):
        val = (v[i] + v[i + 1]) / 2
        a[i] = np.argwhere(v1 > val)[0] - 1
        b[i] = np.argwhere(v2 > val)[0] - 1
    # Last element is always itself in new weights
    a[-1] = len(w1) - 1
    b[-1] = len(w2) - 1
    A_sq = A[a]
    A_sq = A_sq[:, a]
    B_sq = B[b]
    B_sq = B_sq[:, b]
    C = (A_sq - B_sq)

    # Normalize C according to weights
    C = C * (np.outer(w, w))
    return w, C


def _compute_C_eqdim_unweighted(A, B):
    """
    Generates the difference matrix of the two equal dimension unweighted matrices

    Args:
        A: ndarray, (n, n) matrix
        B: ndarray, (m, m) matrix
    Returns:
        C: ndarray, the difference matrix
    """
    n, n2 = np.shape(A)
    n1 = n
    w = np.ones(n) / n
    C = (A - B) / (n1 * n1)  # Normalized C
    return w, C


def _compute_C_uneqdim_unweighted(A, B):
    """
    Generates the difference matrix of the two equal dimension unweighted matrices

    Args:
        A: ndarray, (n, n) matrix
        B: ndarray, (m, m) matrix
    Returns:
        C: ndarray, the difference matrix
    """
    n, n2 = np.shape(A)
    m, m2 = np.shape(B)
    d = math.gcd(n, m)
    k = n // d
    l = m // d
    c = k + l - 1
    v1 = np.arange(k) / n
    v2 = np.arange(1, l + 1) / m
    v = np.hstack((v1, v2))
    np.sort(v)
    w = np.diff(v)
    w = np.tile(w, d)

    # Create matrix of differences
    n1 = len(w)
    vals = np.tile(v[:-1], d) + np.floor(np.arange(n1) / c) / d + 1. / (2 * n1)
    a = np.floor(vals * n).astype(int)
    b = np.floor(vals * m).astype(int)
    A_sq = A[a]
    A_sq = A_sq[:, a]
    B_sq = B[b]
    B_sq = B_sq[:, b]
    C = (A_sq - B_sq)

    # Normalize C according to weights
    C = C * (np.outer(w, w))
    return w, C



from setuptools import setup


def readme():
    with open('README.rst') as f:
        return f.read()


setup(
    name='cutnorm',
    version='0.1.10',
    description=
    'Cutnorm approximation via Gaussian Rounding and Optimization with Orthogonality Constraints',
    long_description=readme(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    url='https://github.com/pingkoc/cutnorm',
    author='Ping-Ko Chiu, Peter Diao, Olewasanmi Koyejo',
    author_email=
    'pchiu5@illinois.edu, peter.z.diao@gmail.com, sanmi@illinois.edu',
    license='MIT',
    packages=['cutnorm', 'cutnorm.tools'],
    install_requires=[
        'numpy>=1.16.0',
        'scikit-learn>=0.22.0',
        'scipy>=1.3.0',
        'numba>=0.45.0'
    ],
    python_requires='>=3.6',
    include_package_data=True,
    zip_safe=False)



"""
Implementation of DBF statistic

from "Distance-based analysis of variance: approximate inference and an application to genome-wide association studies", Christopher Minas and Giovanni Montana
"""

import numpy as np
from sklearn.preprocessing import LabelBinarizer
from scipy.stats import pearson3
from scipy.stats import gamma


def distance_variability(dmatrix, Ic):
    '''
    fstat, trW, trB = distance_variability(dmatrix, Ic)

    compute
    (i) within group variability,
    (ii) between group variability
    (iii) total variability
    (iv) distance based F statistic (DBF)

    Args:
        dmatrix: N by N array of distances
        Ic: N  by G array group membership

    Returns:
        fstat: DBF distance based F statistic, between vs. within group variability
        trW: Within group variability
        trB: Between group variability
    '''
    N = float(dmatrix.shape[0])  # total number of elements
    Ng = np.array(
        (Ic).sum(axis=0),
        dtype=float)  # compute number of elemets in each group

    # Compute the H_c
    Hc = np.dot(Ic * (1.0 / Ng), Ic.T) - (1.0 / N)

    # Compute the Gower's centered product matrix
    Gdel = -0.5 * dmatrix * dmatrix
    #As = np.eye(len(Gs)) - (1.0/N)
    #Gdel = np.dot(np.dot(As, Gs), As)
    Gdel -= Gdel.mean(axis=0)[None, :]
    Gdel -= Gdel.mean(axis=1)[:, None]

    trB = np.trace(np.dot(Hc, Gdel))
    Hc1 = np.eye(len(Hc)) - Hc
    trW = np.trace(np.dot(Hc1, Gdel))

    # Compute the dbf statistic
    fstat = trB / trW

    return fstat, trW, trB


def distribution_parameters(dmatrix, Ic):
    '''
    mc, vc, gc  = distribution_parameters(dmatrix, Ic)
    compute (pearson III) null distribution parameters

    Args:
        dmatrix: N by N array of distances
        Ic: N  by G array group membership

    Returns:
        mc: mean
        vc: variance
        gc: skewness
    '''

    # Compute the test statistic
    N = float(dmatrix.shape[0])  # total number of elements
    n = N  #for convenience
    Ng = np.array(
        (Ic).sum(axis=0),
        dtype=float)  # compute number of elemets in each group

    # Compute the Gower's centered product matrix
    Gdel = -0.5 * dmatrix * dmatrix
    Gdel -= Gdel.mean(axis=0)[None, :]
    Gdel -= Gdel.mean(axis=1)[:, None]

    # Compute the H_c
    Hc = np.dot(Ic * (1.0 / Ng), Ic.T) - (1.0 / N)

    # compute sub-parameters
    A1 = np.diag(Hc).sum()  # trace(Hc)
    A2 = (Hc * Hc).sum()  # Trace(Hc Hc) matrix-wise
    A3 = np.diag(Hc * Hc).sum()  # trace(Hc^2) element-wise
    B1 = np.diag(Gdel).sum()  # tr(Gdel)
    B2 = (Gdel * Gdel).sum()  # Tr(Gdel Gdel) matrix-wise
    B3 = np.diag(Gdel * Gdel).sum()  # trace(Gdel^2) element-wise

    # compute first moment / mean
    mom1 = A1 * B1 / (N - 1)
    mc = mom1

    p11 = ((n - 1) * A3 * B3 + (B1 * B1 - B3) * (A1 * A1 - A3) + 2 *
           (A2 - A3) * (B2 - B3) + 4 * A3 * B3) / (n * (n - 1))
    p22 = (4 * (n - 3) * (2 * A3 - A2) * (2 * B3 - B2) + 2 *
           (2 * A3 - A1 * A1) * (2 * B3 - B1 * B1) * (n - 3) +
           (2 * A2 + A1 * A1 - 6 * A3) *
           (2 * B2 + B1 * B1 - 6 * B3)) / (n * (n - 1) * (n - 2) * (n - 3))
    answermom2 = p11 + p22
    answervc = answermom2 - mc**2

    # compute second moment
    vc1 = (2 * ((N - 1) * A2 - A1 * A1) * ((N - 1) * B2 - B1 * B1))
    vc1 /= (N - 1) * (N - 1) * (N + 1) * (N - 2)
    vc2 = N * (N + 1) * B3 - (N - 1) * (B1 * B1 + 2 * B2)
    vc2 /= (N + 1) * N * (N - 1) * (N - 2) * (N - 3)
    vc2 *= N * (N + 1) * A3 - (N - 1) * (A1 * A1 + 2 * A2)
    mom2 = vc1 + vc2
    # compute var from mom2
    vc = mom2 - mc**2

    # compute subparameters for skew
    A4 = np.diag(np.dot(Hc, np.dot(Hc, Hc))).sum()  # trace(Hc Hc Hc)
    A5 = (np.diag(Hc)**3).sum()  # trace (Hc^3) elementwise
    A6 = (Hc**3).sum()  # sum of elements of Hc^3 elementwise
    A7 = np.dot(np.diag(Hc).T, np.diag(np.dot(Hc, Hc)))
    A8 = np.dot(np.dot(np.diag(Hc).T, Hc), np.diag(Hc))

    B4 = np.diag(np.dot(Gdel, np.dot(Gdel,
                                     Gdel))).sum()  # trace(Gdel Gdel Gdel)
    B5 = (np.diag(Gdel)**3).sum()  # trace (Gdel^3) elementwise
    B6 = (Gdel**3).sum()  # sum of elements of Gdel^3 elementwise
    B7 = np.dot(np.diag(Gdel).T, np.diag(np.dot(Gdel, Gdel)))
    B8 = np.dot(np.dot(np.diag(Gdel).T, Gdel), np.diag(Gdel))

    N2 = N * N
    N3 = N2 * N
    N4 = N3 * N

    n2 = N2
    n3 = N3
    n4 = N4

    answermom3 = (
        (n2 * (n + 1) * (n2 + 15 * n - 4) * A5 * B5 + 4 *
         (n4 - 8 * n3 + 19 * n2 - 4 * n - 16) * A6 * B6 + 24 * (n2 - n - 4) *
         (A6 * B8 + B6 * A8) + 6 *
         (n4 - 8 * n3 + 21 * n2 - 6 * n - 24) * A8 * B8 + 12 *
         (n4 - n3 - 8 * n2 + 36 * n - 48) * A7 * B7 + 12 *
         (n3 - 2 * n2 + 9 * n - 12) * (A1 * A3 * B7 + A7 * B1 * B3) + 3 *
         (n4 - 4 * n3 - 2 * n2 + 9 * n - 12) * A1 * B1 * A3 * B3 + 24 *
         ((n3 - 3 * n2 - 2 * n + 8) * (A7 * B6 + A6 * B7) +
          (n3 - 2 * n2 - 3 * n + 12) *
          (A7 * B8 + A8 * B7)) + 12 * (n2 - n + 4) *
         (A1 * A3 * B6 + B1 * B3 * A6) + 6 * (2 * n3 - 7 * n2 - 3 * n + 12) *
         (A1 * A3 * B8 + A8 * B1 * B3) - 2 * n * (n - 1) * (n2 - n + 4) *
         ((2 * A6 + 3 * A8) * B5 + (2 * B6 + 3 * B8) * A5) - 3 * n * (n - 1) *
         (n - 1) * (n + 4) * ((A1 * A3 + 4 * A7) * B5 +
                              (B1 * B3 + 4 * B7) * A5) + 2 * n * (n - 1) *
         (n - 2) * ((A1 * A1 * A1 + 6 * A1 * A2 + 8 * A4) * B5 +
                    (B1 * B1 * B1 + 6 * B1 * B2 + 8 * B4) * A5) +
         (A1 * A1 * A1) * ((n3 - 9 * n2 + 23 * n - 14) * (B1 * B1 * B1) + 6 *
                           (n - 4) * B1 * B2 + 8 * B4) + 6 * A1 * A2 *
         ((n - 4) * (B1 * B1 * B1) +
          (n3 - 9 * n2 + 24 * n - 14) * B1 * B2 + 4 * (n - 3) * B4) + 8 * A4 *
         ((B1 * B1 * B1) + 3 * (n - 3) * B1 * B2 +
          (n3 - 9 * n2 + 26 * n - 22) * B4) - 16 * ((A1 * A1 * A1) * B6 + A6 *
                                                    (B1 * B1 * B1)) - 6 *
         (A1 * A2 * B6 + A6 * B1 * B2) * (2 * n2 - 10 * n + 16) - 8 *
         (A4 * B6 + A6 * B4) * (3 * n2 - 15 * n + 16) -
         ((A1 * A1 * A1) * B8 + A8 *
          (B1 * B1 * B1)) * (6 * n2 - 30 * n + 24) - 6 *
         (A1 * A2 * B8 + A8 * B1 * B2) * (4 * n2 - 20 * n + 24) - 8 *
         (A4 * B8 + A8 * B4) * (3 * n2 - 15 * n + 24) - (n - 2) *
         (24 * ((A1 * A1 * A1) * B7 + A7 *
                (B1 * B1 * B1)) + 6 * (A1 * A2 * B7 + A7 * B1 * B2) *
          (2 * n2 - 10 * n + 24) + 8 * (A4 * B7 + A7 * B4) *
          (3 * n2 - 15 * n + 24) + (3 * n2 - 15 * n + 6) *
          ((A1 * A1 * A1) * B1 * B3 + A1 * A3 *
           (B1 * B1 * B1)) + 6 * (A1 * A2 * B1 * B3 + A1 * A3 * B1 * B2) *
          (n2 - 5 * n + 6) + 48 * (A4 * B1 * B3 + A1 * A3 * B4))) /
        (n * (n - 1) * (n - 2) * (n - 3) * (n - 4) * (n - 5)))

    answerskew = (answermom3 - 3 * mc * answervc - mc**3) / (answervc**
                                                             (3.0 / 2.0))

    return mc, answervc, answerskew


def inv_f_fn(mc, vc, gc, fstat, trT):
    '''
    inverse fn

    Args:
        mc: mean
        vc: variance
        gc: skewness
        fstat: DBF distance based F statistic, between vs. within group variability
        trT: total group varaibility for DBF statistic

    Returns:
        inv fn val
    '''
    ans = ((trT - mc) * fstat - mc) / (((vc)**(0.5)) * (1 + fstat))
    return (ans)


def dbf_pvalue(mc, vc, gc, fstat, trW, trB):
    '''
    pval = pearson_three(mc, vc, gc, fstat)
    compute one sided p value from standardized pearson three distribution

    Args:
        mc: mean
        vc: variance
        gc: skewness
        fstat: DBF distance based F statistic, between vs. within group variability
        trW: within group variability for DBF statistic
        trB: between group varaibility for DBF statistic

    Returns:
        pval = one sided p value
    '''
    trT = trW + trB  #compute total variability of data

    beta = (trT - mc) / vc  #discontinuity point in trB space
    #alpha = (gc*mc - 2*vc)
    #alpha /= gc*(trT - mc) + 2*vc #endpoint of support in fspace

    alpha = 1.0 / ((gc * trT) / (gc * mc - 2.0 * (vc)**(0.5)) - 1.0)

    a = fstat

    if gc >= 0:
        a1 = gamma.cdf(
            ((trT - mc) / (vc)**(0.5)) + 2 / gc, (4 / (gc**2)), scale=(gc / 2))
        if a >= alpha:
            ans = 1 + gamma.cdf(
                inv_f_fn(mc, vc, gc, fstat, trT) + 2 / gc, (4 / (gc**2)),
                scale=(gc / 2)) - a1
        if a <= -1:
            ans = gamma.cdf(
                inv_f_fn(mc, vc, gc, fstat, trT) + 2 / gc, (4 / (gc**2)),
                scale=(gc / 2)) - a1
        if (-1 < a) and (a < alpha):
            ans = 1 + gamma.cdf(
                inv_f_fn(mc, vc, gc, fstat, trT) + 2 / gc, (4 / (gc**2)),
                scale=(gc / 2)) - a1
    if (gc < 0) and (alpha < -1):
        a1 = gamma.cdf(
            (2 / abs(gc)) - ((trT - mc) / (vc)**(0.5)), (4 / (gc**2)),
            scale=(abs(gc) / 2))
        a1 = 1 - a1
        if a <= alpha:
            ans = gamma.cdf(
                (2 / abs(gc)) - inv_f_fn(mc, vc, gc, fstat, trT),
                (4 / (gc**2)),
                scale=(abs(gc) / 2)) - a1
            ans = 1 - ans
        if a > -1:
            ans = 1 + gamma.cdf(
                (2 / abs(gc)) - inv_f_fn(mc, vc, gc, fstat, trT),
                (4 / (gc**2)),
                scale=(abs(gc) / 2)) - a1
            ans = 1 - ans
        if (alpha < a) and (a <= -1):
            ans = 1 + gamma.cdf(
                (2 / abs(gc)) - inv_f_fn(mc, vc, gc, -1, trT), (4 / (gc**2)),
                scale=(abs(gc) / 2)) - a1
            ans = 1 - ans
    if (gc < 0) and (alpha > -1):
        a1 = gamma.cdf(
            (2 / abs(gc)) - ((trT - mc) / (vc)**(0.5)), (4 / (gc**2)),
            scale=(abs(gc) / 2))
        a1 = 1 - a1
        if a < -1:
            ans = 0
        if (-1 < a) and (a <= alpha):
            ans = gamma.cdf(
                (2 / abs(gc)) - inv_f_fn(mc, vc, gc, fstat, trT),
                (4 / (gc**2)),
                scale=(abs(gc) / 2))
            ans = 1 - ans
        if a > alpha:
            ans = 1
    return (1 - ans)


def dbf_test(dmatrix, labels):
    '''
    pval, fstat, Bvar, Wvar = dbf_test(dmatrix, labels)
    run dbf test

    Args:
        dmatrix: N by N array of distances
        Labels: N array group membership

    Returns:
        pval: one sided p value
        fstat: DBF distance based F statistic, between vs. within group variability
        Bvar: between vs overall group variability
        Wvar: within vs overall group variability
    '''
    G = len(np.unique(labels))
    N = len(labels)
    # check at least 2 grroups
    assert G >= 2, 'Group variability with less than 2 groups not allowed, input G=%d' % (
        G, )
    # check at least 5 samples
    assert N >= 5, 'Test statistic requires N >=5, input N=%d' % (N, )

    # create label binarizer object
    lb = LabelBinarizer()
    # form N x G group label matrix
    # NOTE: when there are only two groups, label binarizer only returns a one column matrix.
    # TEMP FIX: check if G==2, and ad an extra column
    Ic = np.array(lb.fit_transform(labels), dtype=float)
    if G < 3:  # G==2
        Ic = np.hstack((Ic, 1.0 - Ic))

    # compute test statistic
    fstat, trW, trB = distance_variability(dmatrix, Ic)
    Bvar = trB / (trW + trB)  # between vs overall group variability
    Wvar = trW / (trW + trB)  # within vs overall group variability

    # compuete Null parameters
    mc, vc, gc = distribution_parameters(dmatrix, Ic)

    # evaluate P value
    pval = dbf_pvalue(mc, vc, gc, fstat, trW, trB)

    return pval, fstat, Bvar, Wvar
